%FindInducedPower Combines force data with kinematics from image processing
%   to estimate induced power during perch to perch flight
%
%   Required: getTimings.m, getScaling.m, and .mat flight data files
clear; clc; %close all;

flight = 'Gaga5'; 
recordingDate = '11-7';

addpath('PlateData');   % location of force data files
addpath('AngleData');   % location of .mat files from ProcessBottomView.m

% Get force and kinematic data
fileName = [recordingDate '-' flight];
load([fileName '_plates']);
load(fileName);

% Get wingbeat timing and relative timing of video and force sensors
[flightType, sensorTime, dstrokeStart, dstrokeEnd, camTime] = getTimings(fileName);
fps = 3000;     % frames per second of high speed video
offset = sensorTime + camTime/fps;  % timing offset (in seconds) 

% Synchronize data
if offset > 0
    forcesBegin = round((min(time) + offset)*1000);
    forcesEnd = min(length(Plates), round((max(time) + offset)*1000));
    Fz = Plates(forcesBegin:2:forcesEnd); % Lift during flight
else
    forcesBegin = round(min(time)*1000-offset*fps/tskip);
    forcesEnd = min(length(Plates),forcesBegin+2*length(time)-1); %round(max(time)*1000);
    Fz = Plates(forcesBegin:2:forcesEnd); 
end
angle_s = angle_s(1:length(Fz));
r_s = r_s(1:length(Fz));
time = time(1:length(Fz)); 

% Convert pixel lengths to meters
[ scaleStart, scaleEnd ] = getScaling( recordingDate, flightType );
scale = linspace(scaleStart, scaleEnd, length(r_s))'; 
l = scale.*r_s;              % estimated length of wing (m) 

% Find average L, r, and change in stroke plane angle for each downstroke
strokes = length(dstrokeStart);
L = zeros(1, strokes);    % average lift during downstroke
r = zeros(1, strokes);    % average wing length
phi = zeros(1, strokes);  % angle swept out by each wing
for i = 1:strokes    
    index1 = find(time <= dstrokeStart(i), 1, 'last');
    index2 = find(time <= dstrokeEnd(i), 1, 'last');
    L(i) = mean(Fz(index1:index2));
    r(i) = mean(l(index1:index2));
    phi(i) = max(angle_s(index1:index2)) - min(angle_s(index1:index2));
end

% Calculate induced power for each downstroke using actuator disc theory
Ad = pi*r.^2.*phi/180*cosd(11)*cosd(11);   % estimated disc area (m^2)
rho = 1.18;                     % density of air at 25 deg C (kg/m^3)
P = sqrt((L.^3)./(2*rho.*Ad));  % induced power


% Repeat process for upstrokes
L_up = zeros(1, strokes);    % average lift during upstroke
r_up = zeros(1, strokes);    % average wing length
phi_up = zeros(1, strokes);  % angle swept out by each wing
for i = 1:strokes
    index1 = find(time >= dstrokeEnd(i), 1, 'first');
    if i < strokes
        index2 = find(time <= dstrokeStart(i+1), 1, 'last');
    else
        index2 = find(time <= dstrokeEnd(i) + 0.02, 1, 'last');
    end
    L_up(i) = mean(Fz(index1:index2));
    r_up(i) = mean(l(index1:index2));
    phi_up(i) = max(angle_s(index1:index2)) - min(angle_s(index1:index2));
end
Ad_up = pi*r_up.^2.*phi_up/180*cosd(11)*cosd(11);   % estimated disc area (m^2)
P_up = sqrt((L_up.^3)./(2*rho.*Ad_up));  % induced power



% ----------------- Plot results ---------------------
tmax = dstrokeEnd(end)+.05; warning off;
figure;
% plot forces with wingbeats
subplot('position', [0.1 0.68 0.8 0.275]); hold on; 
for k = 1:length(dstrokeEnd)       
    h = area([dstrokeStart(k) dstrokeEnd(k)], [1 1], -.1,...
        'FaceColor', [.8, .8, .8], 'EdgeColor', 'none');
end
plot(time, Fz, 'k', 'LineWidth', 2); grid on;
legend(h, 'Downstroke', 'Location', 'NorthEast');
ylabel('Lift [N]'); axis([0 tmax -.1 1]); title('Flight Analysis Results'); 
set(gca, 'XTickLabels', {'' '0.5' '1' '1.5' '2' ''}, 'FontSize', 18); 

% plot stroke angle with wingbeats
subplot('position', [0.1 0.38 0.8 0.295]); hold on; 
for k = 1:length(dstrokeEnd)       
    h = area([dstrokeStart(k) dstrokeEnd(k)], 190*[1 1], -10,...
        'FaceColor', [.8, .8, .8], 'EdgeColor', 'none');
end
plot(time, angle_s, 'b', 'LineWidth', 2); %xlabel('Time [s]');
ylabel('Stroke Angle [deg]'); axis([0 tmax -10 190]); grid on;
set(gca, 'XTickLabels', {'' '0.5' '1' '1.5' '2' ''}, 'FontSize', 18); 

% plot induced power for each wingbeat
subplot('position', [0.1 0.08 0.8 0.295]); hold on; 
x = .5*(dstrokeStart+dstrokeEnd);
x_up = .5*(dstrokeStart(2:end) + dstrokeEnd(1:end-1));
x_up = [x_up dstrokeEnd(end)+.01];
h_down = bar(x, P, .3);
h_up = bar(x_up, P_up, .3);
axis([0 tmax 0 5.5]);
xlabel('Time [s]'); ylabel('Induced Power [W]');
h_down.FaceColor = [.8 0 0]; h_up.FaceColor = [.5 .5 1]; grid on;
legend([h_down, h_up], {'Downstroke', 'Upstroke'}, 'Location', 'NorthEast');
set(gca,'FontSize',18); %,'FontName','Times','FontWeight','Normal', 'XTickLabel', xlabels);

