% ProcessBottomView finds the time-resolved stroke angle during a
% perch-to-perch flight using a video showing the bottom view of the flight

% Functions that may be needed (depending on selected options): 
%   getParameters.m, getFrames.m, getRegion.m, getScaling.m 
%   makeVideo.m, combineVideos.m, showVideoWithPlot.m

clc, clear, close all

% ------ Flight to process -----
flight = 'Gaga5'; 
recordingDate = '11-7';

%addpath(['../' recordingDate '-15']);   % location of video
videoName = [flight '_13503.avi'];      % name of video for processing

% ------ Options -----
CREATE_VIDEO = 0;   % 1 - create video showing stroke angle w/ flight
SHOW_PLOT = 1;      % 1 - plot stroke angle and wingbeats

SAVE_DATA = 1;      % 1 - save final stroke angle vs. time data 
                    %     along with estimated wing radius as a .mat file

SELECT_REGION = 0;  % 1 - manually select region to analyze
EXISTING_FILE = 1;  % 1 - file exists in getParameters (if 0, processing
                    %     parameters must be specified in the next section)

% ------------Get flight-specific parameters -------------
if (EXISTING_FILE)
    [gamma, thresh, Rthresh, range1, range2a, range2b, range3a, range3b, range4, ...
    search, smoothAngle, initialDS, NP, flightType] = getParameters(flight, recordingDate);
else
    gamma = 1/3;        % for correcting contrast of frames
    thresh = .05;       % threshold for binarizing frames
    range1 = [0,120];   % stroke angle limits for first section of flight
    Rthresh = 50;       % min distance to wingtip with preferred angle limits
    range2a = [0, 120]; % preferred limits for second section
    range2b = [10,120]; % alternative limits for second section
    range3a = [20,180]; % preferred limits for third section
    range3b = [20,180]; % alternative limits for third section
    range4 = [40, 170]; % angle limits for final section of flight
    search = [15, 250]; % frames to search for peaks
    smoothAngle = .05;  % smoothing level for stroke angle
    initialDS = 45;     % threshold for finding initial downstroke time
    NP = 10;            % max number of peaks to identify
    flightType = 'Level';
end

%% -----------------------------------------------------------
%                 Begin Video Processing                       
%  -----------------------------------------------------------
% Settings
res = 0.8;          % scale images by res 
tskip = 6;          % save every tskip images 
fps = 3000;         % FPS of original video

% Get video frames
[ origFrame, nFrames ] = getFrames( videoName, res, tskip );

%%
% Adjust contrast and suppress background 
Frame = cell(nFrames,1);
for i = 1:nFrames
    disp(['Correcting frame ' sprintf('%d/%d', i,nFrames) ' of ' videoName '...']);
    Frame{i} = real(origFrame{i}.^gamma - .5*origFrame{nFrames}.^gamma - .5*origFrame{1}.^gamma); 
    clc
end
%%

% Get region of original frame to use for analysis
if(SELECT_REGION)
    % Create time-lapse image using 10 images
    showFrame = Frame{1};
    for i = 1:10   
        showFrame = showFrame + Frame{round(i/10*nFrames)};
    end
    
    % Chose region to analyze in frames 
    disp('Choose region showing only complete bottom view')
    [ region, xmin, ymin ] = getRegion( showFrame, Frame );
    clc;
    
else
    disp('Creating region to analyze...');   
    if strcmp(flightType, 'Ascend')
        if strcmp(recordingDate, '10-25')
            xmin = 320; xmax = 810;
        else    
            xmin = 222; xmax = 730;
        end
        ymin = 235; ymax = 410; 
    elseif strcmp(flightType, 'Descend')
        xmin = 250; xmax = 730;
        ymin = 200; ymax = 410;
    elseif strcmp(flightType, 'Level')
        if strcmp(recordingDate, '10-25')
            xmin = 320; xmax = 800;
        else        
            xmin = 225; xmax = 730;
        end
        ymin = 190; ymax = 410;        
    end
    N = length(Frame);
    region = cell(N,1);
    for i=1:N   
        region{i} = Frame{i}(ymin:ymax,xmin:xmax);    
    end
end


% Binarize image
%thresh = 0.04;
FrameBW = region;
for i = 1:nFrames
    % remove unwanted white regions
    FrameBW{i} = imopen(im2bw(FrameBW{i}, thresh), strel('square',10));
    % remove unwanted black regions
    FrameBW{i} = imclose(FrameBW{i}, strel('square',7));
end
%imshow([FrameBW{20} FrameBW{70}; FrameBW{150} FrameBW{200}])

%%
% Find centroids of binarized regions
disp('Finding centroids...');
centroids = cell(1,nFrames);
startFrame0 = 1; stopFrame0 = nFrames;
s = regionprops(FrameBW{startFrame0}, 'centroid');
while isempty(s) % in case no region is detected in current frame
    startFrame0 = startFrame0 + 1;
    s = regionprops(FrameBW{startFrame0}, 'centroid');
end
for i = startFrame0:nFrames
    s = regionprops(FrameBW{i}, 'centroid');
    if ~isempty(s)
        centroids{i} = s.Centroid;
    else
        if i<nFrames && isempty(regionprops(FrameBW{i+1}, 'centroid'))
            stopFrame0 = i-1; break;
        elseif ~isempty(regionprops(FrameBW{i+1}, 'centroid'))
            % remove frame
            for j = i:nFrames-1
                Frame{j} = Frame{j+1};
                FrameBW{j} = FrameBW{j+1};
            end
            centroids{i} = centroids{i-1};
            nFrames = nFrames - 1;
            stopFrame0 = nFrames;
            i = i-1;
        end
    end
end

% Get centroid coordinates in original frame
x0 = centroids{startFrame0}(:,1)+xmin; 
y0 = centroids{startFrame0}(:,2)+ymin;
x_raw = x0*ones(nFrames,1); y_raw = y0*ones(nFrames,1);
for i = startFrame0:stopFrame0
    x_raw(i) = centroids{i}(:,1)+xmin;
    y_raw(i) = centroids{i}(:,2)+ymin;
end

% Leave out frames where the bird is not moving 
ymean = mean(y_raw);
dy = abs(y_raw-ymean);
startFrame = startFrame0+find(dy(startFrame0:startFrame0+25)>5, 1, 'last');
if isempty(startFrame), startFrame = startFrame0; end
stopFrame = stopFrame0-25+find(dy(stopFrame0-25:stopFrame0)>5, 1, 'first');
if isempty(stopFrame) || stopFrame > stopFrame0, stopFrame = stopFrame0; end

% Smooth coordinates
x_c = smooth(x_raw(startFrame:stopFrame), 0.7, 'loess');
y_c = smooth(y_raw(startFrame:stopFrame), 0.7, 'loess');
%plot(x_c, y_c)
%%
% Use centroid path to shift video frames so bird remains centered
disp('Shifting frames...');
[height, width] = size(Frame{1});
shifted = Frame;
xmean = round(mean(x_c));
ymean = round(mean(y_c));
for i = 1:length(x_c)
    disp(['Shifting frame ' sprintf('%d/%d', i,length(x_c)) '...']);    
	dx = xmean - x_c(i); % horizontal shift in pixels
    dy = ymean - y_c(i); % vertical shift in pixels
    A = [1 0 dx; 0 1 dy; 0 0 1];
    tform = maketform('affine', A.'); 
    j = i+startFrame-1;
    shifted{j} = imtransform(Frame{j}, tform, 'bilinear', ...
        'XData', [1 width], 'YData', [1 height], 'FillValues', 0);
    clc;
end
makeVideo('Shifted', shifted, 1, length(x_c));
%% Create new region for analysis
N = length(shifted);
region2 = cell(N,1);
if strcmp(flightType, 'level')
    width = 180; height = 200; % for level flight
else
    width = 200; height = 250; % for ascending/descending flight
end
xmin = xmean-width/2; xmax = xmean+width/2;
ymin = ymean-height/2; ymax = ymean+height/2;
for i=1:N   
    region2{i} = shifted{i}(ymin:ymax,xmin:xmax);    
end

% Binarize image again and find edges
disp('Finding edges...');
FrameBW = region2;
for i = 1:nFrames
    FrameBW{i} = imopen(im2bw(FrameBW{i}, thresh), strel('square',5));
    FrameBW{i} = imclose(FrameBW{i}, strel('square',5));
end
start = startFrame;
stop = stopFrame;
N = stop-start;         % number of frames to analyze
edges = cell(1, N);
for i = 1:N
    outer = imdilate(FrameBW{i+start-1},strel('square', 3));
    edges{i} = outer - FrameBW{i+start-1};
end

%% Find wingtip
disp('Locating wingtips...');
shoulderX = width/2; 
shoulderY = height/2 + 3;
wingtip = zeros(N, 2);  % coordinates of wingtip in edge frames
angle = zeros(N, 1);    % estimated stroke angle (in degrees)
r = zeros(N,1);         % estimated distance from wingtip to shoulder 
angle(1) = 10;
wingtip(1,:) = [shoulderX, shoulderY];
for i = 2:N
    p = regionprops(edges{i}, 'PixelList'); % get position of edge pixels
    pixels= p.PixelList; 
    
    x = pixels(:,1);
    y = pixels(:,2);
    
    % convert to polar coordinates
    R = sqrt((x-shoulderX).^2 + (y-shoulderY).^2);
    theta = atan2d(y-shoulderY,x-shoulderX);
    
    % find point farthest from shoulder
    [Rsorted, index] = sort(R, 'descend');
    
    % Only choose points on right wing 
    z = 1;
    if i < N/8      % For initial part of video (usually pre-takeoff)
        while z < length(index)
            if theta(index(z)) > range1(1) && theta(index(z)) < range1(2)
                break;
            end
            z = z+1;
        end                
    elseif i < N/3      % For first main section of flight
        while z < length(index)
            if R(index(z)) > Rthresh 
                if theta(index(z)) > range2a(1) && theta(index(z)) < range2a(2)
                    break; 
                end
            end
            z = z + 1;
        end
        if z == length(index)
            z = 1;
            while theta(index(z)) < range2b(1) || theta(index(z)) > range2b(2)
                z = z+1;
            end  
        end
    elseif i < (2*N/3)  % For middle section of flight
        while z < length(index)
             if R(index(z)) > Rthresh 
                if theta(index(z)) > range3a(1) && theta(index(z)) < range3a(2)
                    break; 
                end
            end                   
            z = z + 1;
        end
        if z == length(index)
            z = 1;
            while theta(index(z)) < range3b(1) || theta(index(z)) > range3b(2)
                z = z+1;
            end  
        end
    elseif i >= (2*N/3)     % For final section of flight
        while z < length(index)
            if theta(index(z)) > range4(1) && theta(index(z)) < range4(2)
                break;
            end
            z = z+1;
        end                
    end

    % Convert back to cartesian coordinates and limit angles to 0-180 deg
    X = R(index(z))*cosd(theta(index(z)))+shoulderX; 
    Y = R(index(z))*sind(theta(index(z)))+shoulderY;
    wingtip(i,:) = [X, Y];
    if theta(index(z)) < 0 && theta(index(z)) > -15
        angle(i) = 0;
    elseif theta(index(z)) < -165 && theta(index(z)) > -180
        angle(i) = 180;
    else        
        angle(i) = theta(index(z));
    end
    r(i) = R(index(z));
end


%% Identify wingbeats
disp('Identifying wingbeats...');
%NP = 9;
time = 1/fps*tskip*(startFrame:startFrame+length(angle)-1);
startSearch = search(1);
stopSearch = search(2);
angle_s = smooth(angle, smoothAngle, 'loess');  % smooth angle trace
r_s = smooth(r, .07, 'loess');                  % smooth wing radius

% find peaks in angles (for potential start of downstrokes)
[~, locs] = findpeaks(angle_s(startSearch:stopSearch), 'MinPeakProminence', 15, 'NPeaks', NP, 'SortStr', 'descend');
locs = sort(locs) + startSearch;
for i = 1:length(locs) % avoid including bounds
    while angle_s(locs(i)) > 170 
        locs(i) = locs(i) + 1;
    end
    while abs(angle_s(locs(i)) - angle_s(locs(i)+3)) < 10 
        locs(i) = locs(i) + 1;
    end
end
DstrokeStart = (locs+startFrame-1)/fps*tskip; % convert from frames to time

% find dips in angles (for potential end of downstrokes)
[~, locs] = findpeaks(-angle_s(locs(1):stopSearch), 'MinPeakProminence', 5, 'NPeaks', length(DstrokeStart), 'SortStr', 'descend');
locs = sort(locs);
DstrokeEnd = (locs-5)/fps*tskip + DstrokeStart(1);

% adjust first downstroke start time if angle actually increases 
DstrokeStart(1) = (find(angle(1:locs(1)) < initialDS, 1, 'last')+startFrame-1)/fps*tskip;

l_start = length(DstrokeStart);
l_end = length(DstrokeEnd);
% remove any extra mins
k = 1;
while k <= l_end
    if DstrokeEnd(k) < DstrokeStart(k)
        DstrokeEnd(k:l_end-1) = DstrokeEnd(k+1:l_end);        
        l_end = l_end-1;
        DstrokeEnd = DstrokeEnd(1:end-1);
    end
    k = k+1;
end

% remove any extra maxes
k = 1;
while k <= min(l_start-1, l_end)
    if DstrokeEnd(k) > DstrokeStart(k+1)
        DstrokeStart(k:l_start-1) = DstrokeStart(k+1:l_start);
        DstrokeStart = DstrokeStart(1:end-1);
        l_start = l_start - 1;            
    end
    k = k+1;
end

% remove extra strokes
for k = 1:min(l_start, l_end)
    if (DstrokeEnd(k) - DstrokeStart(k)) < 0.007
        DstrokeStart(k:l_start-1) = DstrokeStart(k+1:l_start);
        DstrokeEnd(k:l_end-1) = DstrokeEnd(k+1:l_end);
        l_start = l_start-1; 
        l_end = l_end - 1;
        break;
    end
end

DstrokeStart = DstrokeStart(1:l_start);
DstrokeEnd = DstrokeEnd(1:l_end);

% -----------------------------------------------------------
%                       Show Results                       
%  -----------------------------------------------------------
if (CREATE_VIDEO)
    % create video marking identified wingtip locations
    makeVideo('test', edges, 1, N, wingtip);
    [wingtips, numFrames] = getFrames('test.avi', 1, 1);

    % cut original video to match number of frames in edge video
    Frame_cut = cell(1,N);
    for f = 1:N
        Frame_cut{f} = region{f+start};
    end
    combineVideos(2, 'combo', Frame_cut, wingtips, search(1), search(2));

    % show combined video with plot of stroke angle
    showVideoWithPlot('combo.avi', time(search(1):search(2)), ...
        angle_s(search(1):search(2)), DstrokeStart, DstrokeEnd,...
        'Time (s)', 'Angle (deg)', 'Stroke Angle',[flight '_strokeAngle.avi']);
end

if (SHOW_PLOT)
    figure; 
    hold on;    % show wingbeats
    for k = 1:min(length(DstrokeEnd), length(DstrokeStart))       
        area([DstrokeStart(k) DstrokeEnd(k)], 180*[1 1], 0,...
            'FaceColor', [.8, .8, .8], 'EdgeColor', 'none');
    end

    plot(time, angle_s, 'LineWidth', 2); 
    axis([0 time(end) 0 180]); xlabel('Time (s)'); ylabel('Angle (deg)');
    title('Estimated Stroke Angle'); set(gca, 'FontSize', 12);
end


if (SAVE_DATA)
    fileName = [recordingDate '-' flight];
    save(fileName, 'angle_s', 'r_s', 'time');
end

